Build Packages / build:rpm (rocky9_sls9) (push) Successful in 18m57s
Build Packages / Unit tests (push) Skipped
Build Packages / build:windows:nocuda (push) Successful in 16m55s
Build Packages / build:windows:cuda (push) Successful in 18m48s
Build Packages / build:viewer-tgz:cpu (push) Successful in 13m10s
Build Packages / build:viewer-tgz:cuda (push) Successful in 14m45s
Build Packages / build:rpm (rocky8_nocuda) (push) Successful in 22m23s
Build Packages / build:rpm (rocky9_nocuda) (push) Successful in 20m12s
Build Packages / build:rpm (ubuntu2204_nocuda) (push) Successful in 23m7s
Build Packages / build:rpm (ubuntu2404_nocuda) (push) Successful in 20m43s
Build Packages / build:rpm (rocky8_sls9) (push) Successful in 23m9s
Build Packages / XDS test (durin plugin) (push) Successful in 12m26s
Build Packages / build:rpm (rocky9) (push) Successful in 24m58s
Build Packages / Generate python client (push) Successful in 50s
Build Packages / build:rpm (ubuntu2404) (push) Successful in 23m20s
Build Packages / Create release (push) Skipped
Build Packages / XDS test (JFJoch plugin) (push) Successful in 12m37s
Build Packages / build:rpm (rocky8) (push) Successful in 27m58s
Build Packages / build:rpm (ubuntu2204) (push) Successful in 25m38s
Build Packages / Build documentation (push) Successful in 59s
Build Packages / DIALS test (push) Successful in 23m16s
Build Packages / XDS test (neggia plugin) (push) Successful in 6m38s
**Files written by Jungfraujoch now import correctly in DIALS, XDS and pyFAI.** A tilted detector, a grid scan, a still recorded at a goniometer position, and saturated or unreadable pixels were each described in a way that a third-party program acted on wrongly. If you process Jungfraujoch data outside Jungfraujoch, prefer this release to any earlier one. * HDF5: the detector tilt (`rot1`/`rot2`/`rot3`) is exported correctly in the NXmx transformation chain; untilted geometries are unaffected. * HDF5: a still recorded at a goniometer position is no longer read back as a single image, and a grid scan records a stationary spindle so a program that requires a rotation axis can open it. * HDF5: the sample transformation chain is written in mounting order, with a Smargon head position told apart from the spindle, one entry per image, `module_offset` as a float unit vector, and `offset_units` on every offset. * HDF5: saturated, underloaded and unreadable pixels are described so a downstream program masks them - `saturation_value`, `underload_value`, `error_value` and `bit_depth_readout` are written correctly, and a data file missing next to a VDS master reads as the error marker rather than as zero counts. * HDF5: the rotation axis is read back under whatever name it carries, and `mirror_y` records whether the assembled image is mirrored in Y relative to the detector's raw readout. * A grid scan and a goniometer axis can both be set; they are no longer alternatives. * `images_per_file` is chosen from the acquisition when it is not given: a rotation sweep of at most 20000 images goes into a single data file, a grid scan splits on whole fast-axis rows, and stills and serial keep 1000. * The writer refuses a stream whose start message declares a different pixel format than its images carry, and a DECTRIS detector sending signed images is no longer declared unsigned. * The image stream can carry the sample transformation chain (`transformations`, in the END message); a producer that does not send it gets the same chain built by the writer. * rugnux: fixing the space group with `-S` no longer prevents the lattice from being found - a lattice indexed in a different setting is reindexed into that group's own setting, and a run whose crystal does not have that group's lattice stops and names the cell it indexed as, rather than reporting statistics that cannot describe it. * rugnux: the per-image resolution estimate now predicts the resolution the merged data reach rather than the highest-resolution spot found, and is reported as `SPOT_RESOLUTION_ESTIMATE`. * rugnux: two runs of the same command on the same images produce the same merged intensities; the azimuthal profile written alongside them is not yet reproducible in the same way. * rugnux: the offline lattice refinement is bounded by iterations rather than by a wall clock, so a loaded machine can no longer refine to a different lattice; a live acquisition keeps its real-time bound. * rugnux: the detector-frame modulation correction is fitted on a grid spanning the detector, so whether it is applied no longer depends on how far integration reached. * rugnux: the geometry pre-pass no longer writes `<prefix>_01.mtz`, `_01.cif`, `_01.hkl` and `_01_image.dat`; the refined second pass writes those files under `<prefix>`, and that is the result to use. * rugnux: `_process.h5` describes the pixel format of the images it links to, and is written on a thread of its own. * rugnux: the detector geometry is also logged in XDS's convention (`ORGX`/`ORGY`, detector axis vectors, rotation axis), so it can be compared with an XDS refinement. * rugnux: an image integrated in pyFAI through the `.poni` file written by `--mode calibration` comes out with the correct azimuth, and the file declares pyFAI's `orientation`, which needs pyFAI 2024.01 or newer. Radial integration is unchanged. * rugnux: a rotation run is substantially faster throughout - beam-stop detection, first-pass indexing, geometry refinement, integration, scaling and merging - and observations outside the scaling resolution range are dropped as they are ingested. The refined geometry, the space group chosen and the merged statistics are unchanged. * Faster spot finding and indexing, on the broker as well as in rugnux; the spots found and the lattices indexed are unchanged. * A run reserves substantially less GPU memory: nothing is allocated for buffers that are never read, and a worker builds only the engines it uses. * rugnux: with `-N` left at its default the per-image loop of `--mode mx` uses at most 16 workers per GPU, rather than one per hardware thread; an explicit `-N` is obeyed as given. * CUDA 12 builds now contain device code for Volta, so the RHEL 8 packages and the portable Linux `.tgz` run on a V100; the CUDA 13 artefacts (RHEL 9, Ubuntu, Windows) remain Turing and newer. * The build resolves a single Eigen for the whole project, and refuses to configure if Ceres picks up a different one; a build that mixed two Eigen versions was undefined behaviour and crashed at -O2. * Documentation: a security page, and the supported GPU generations and minimum NVIDIA driver version of every released artefact. **Breaking change to OpenAPI** - regenerate the client (`jfjoch-client` 1.0.0-rc.162, `frontend/src/client`): * `dataset_settings.images_per_file` is no longer `default: 1000` and no longer accepts `0`; it is optional, and its minimum is 1. A client sending `0` (previously "one file for the whole run") is now rejected - omit the field instead, which for a rotation sweep gives the same single file. * `file_writer_format` now defaults to `NXmxVDS`, matching the server's own default and the layout recommended for DIALS, XDS and CrystFEL. A generated client that fills in schema defaults and does not set the format explicitly will write VDS masters where it previously wrote legacy ones; set `NXmxLegacy` explicitly to keep them. --------- Co-authored-by: jungfrau <jungfrau@mx-aare-test.psi.ch> Reviewed-on: #72 Co-authored-by: Filip Leonarski <filip.leonarski@psi.ch>
162 lines
6.9 KiB
Plaintext
162 lines
6.9 KiB
Plaintext
// SPDX-FileCopyrightText: 2026 Filip Leonarski, Paul Scherrer Institute <filip.leonarski@psi.ch>
|
|
// SPDX-License-Identifier: GPL-3.0-only
|
|
|
|
#include <climits>
|
|
|
|
#include "ROIIntegrationGPU.h"
|
|
#include "../../common/DiffractionExperiment.h"
|
|
|
|
inline void cuda_err(cudaError_t val) {
|
|
if (val != cudaSuccess)
|
|
throw JFJochException(JFJochExceptionCategory::GPUCUDAError, cudaGetErrorString(val));
|
|
}
|
|
|
|
// One pixel carries a 16-bit mask, so it can feed any subset of the ROIs.
|
|
// Each block reduces into shared memory first to keep global atomics low.
|
|
__global__
|
|
void gpu_roi(
|
|
const uint16_t *__restrict__ roi_map,
|
|
const int32_t *__restrict__ input_buffer,
|
|
size_t num_pixels,
|
|
size_t width,
|
|
int roi_count,
|
|
unsigned long long *__restrict__ roi_sum,
|
|
unsigned long long *__restrict__ roi_sum2,
|
|
unsigned long long *__restrict__ roi_pixels,
|
|
unsigned long long *__restrict__ roi_x_weighted,
|
|
unsigned long long *__restrict__ roi_y_weighted,
|
|
int *__restrict__ roi_max) {
|
|
extern __shared__ unsigned long long shared[];
|
|
unsigned long long *s_sum = shared;
|
|
unsigned long long *s_sum2 = &s_sum[roi_count];
|
|
unsigned long long *s_pixels = &s_sum2[roi_count];
|
|
unsigned long long *s_xw = &s_pixels[roi_count];
|
|
unsigned long long *s_yw = &s_xw[roi_count];
|
|
int *s_max = (int *) &s_yw[roi_count];
|
|
|
|
for (int r = threadIdx.x; r < roi_count; r += blockDim.x) {
|
|
s_sum[r] = 0;
|
|
s_sum2[r] = 0;
|
|
s_pixels[r] = 0;
|
|
s_xw[r] = 0;
|
|
s_yw[r] = 0;
|
|
s_max[r] = INT_MIN;
|
|
}
|
|
__syncthreads();
|
|
|
|
for (size_t idx = blockIdx.x * blockDim.x + threadIdx.x;
|
|
idx < num_pixels;
|
|
idx += blockDim.x * gridDim.x) {
|
|
const uint16_t mask = roi_map[idx];
|
|
if (mask == 0)
|
|
continue;
|
|
|
|
const int32_t v = input_buffer[idx];
|
|
if (v == INT32_MIN) // masked/bad pixel
|
|
continue;
|
|
const bool saturated = (v == INT32_MAX);
|
|
|
|
const long long val = v;
|
|
const long long x = idx % width;
|
|
const long long y = idx / width;
|
|
const unsigned long long val_u = (unsigned long long) val;
|
|
const unsigned long long val2_u = (unsigned long long) (val * val);
|
|
const unsigned long long vx_u = (unsigned long long) (val * x);
|
|
const unsigned long long vy_u = (unsigned long long) (val * y);
|
|
|
|
for (int r = 0; r < roi_count; r++) {
|
|
if (!(mask & (1u << r)))
|
|
continue;
|
|
if (!saturated) {
|
|
atomicAdd(&s_sum[r], val_u);
|
|
atomicAdd(&s_sum2[r], val2_u);
|
|
atomicAdd(&s_pixels[r], 1ULL);
|
|
atomicAdd(&s_xw[r], vx_u);
|
|
atomicAdd(&s_yw[r], vy_u);
|
|
}
|
|
atomicMax(&s_max[r], v);
|
|
}
|
|
}
|
|
__syncthreads();
|
|
|
|
for (int r = threadIdx.x; r < roi_count; r += blockDim.x) {
|
|
atomicAdd(&roi_sum[r], s_sum[r]);
|
|
atomicAdd(&roi_sum2[r], s_sum2[r]);
|
|
atomicAdd(&roi_pixels[r], s_pixels[r]);
|
|
atomicAdd(&roi_x_weighted[r], s_xw[r]);
|
|
atomicAdd(&roi_y_weighted[r], s_yw[r]);
|
|
atomicMax(&roi_max[r], s_max[r]);
|
|
}
|
|
}
|
|
|
|
ROIIntegrationGPU::ROIIntegrationGPU(const DiffractionExperiment &experiment, std::shared_ptr<CudaStream> stream)
|
|
: ROIIntegration(experiment),
|
|
stream(stream),
|
|
gpu_roi_map(npixel),
|
|
gpu_sum(roi_count),
|
|
gpu_sum2(roi_count),
|
|
gpu_pixels(roi_count),
|
|
gpu_x_weighted(roi_count),
|
|
gpu_y_weighted(roi_count),
|
|
gpu_max(roi_count),
|
|
host_sum(roi_count),
|
|
host_sum2(roi_count),
|
|
host_pixels(roi_count),
|
|
host_x_weighted(roi_count),
|
|
host_y_weighted(roi_count),
|
|
host_max(roi_count),
|
|
max_init(roi_count, INT_MIN) {
|
|
|
|
// The current device, not device 0: workers are pinned round-robin across the GPUs, so device 0's
|
|
// SM count can belong to a different card than the one these kernels launch on.
|
|
int device = 0;
|
|
cuda_err(cudaGetDevice(&device));
|
|
cudaDeviceProp prop{};
|
|
cuda_err(cudaGetDeviceProperties(&prop, device));
|
|
threads = 128;
|
|
blocks = 4 * prop.multiProcessorCount;
|
|
shared_needed = roi_count * (5 * sizeof(unsigned long long) + sizeof(int));
|
|
|
|
// On this engine's stream, like every other operation it issues: the streams are non-blocking, so a
|
|
// NULL-stream copy is no longer ordered against the kernels that read the map. The one-time
|
|
// synchronise leaves the constructor with the upload settled rather than in flight.
|
|
cuda_err(cudaMemcpyAsync(gpu_roi_map, roi_map.data(), sizeof(uint16_t) * npixel,
|
|
cudaMemcpyHostToDevice, *stream));
|
|
cuda_err(cudaStreamSynchronize(*stream));
|
|
}
|
|
|
|
void ROIIntegrationGPU::Run(const ImagePreprocessorBuffer &image, std::map<std::string, ROIMessage> &out) {
|
|
if (image.size() != npixel)
|
|
throw JFJochException(JFJochExceptionCategory::InputParameterInvalid,
|
|
"ROIIntegration: mismatch in image size");
|
|
|
|
cuda_err(cudaMemsetAsync(gpu_sum, 0, sizeof(unsigned long long) * roi_count, *stream));
|
|
cuda_err(cudaMemsetAsync(gpu_sum2, 0, sizeof(unsigned long long) * roi_count, *stream));
|
|
cuda_err(cudaMemsetAsync(gpu_pixels, 0, sizeof(unsigned long long) * roi_count, *stream));
|
|
cuda_err(cudaMemsetAsync(gpu_x_weighted, 0, sizeof(unsigned long long) * roi_count, *stream));
|
|
cuda_err(cudaMemsetAsync(gpu_y_weighted, 0, sizeof(unsigned long long) * roi_count, *stream));
|
|
cuda_err(cudaMemcpyAsync(gpu_max, max_init.data(), sizeof(int) * roi_count, cudaMemcpyHostToDevice, *stream));
|
|
|
|
gpu_roi<<<blocks, threads, shared_needed, *stream>>>(
|
|
gpu_roi_map, image.getGPUBuffer(), npixel, width, roi_count,
|
|
gpu_sum, gpu_sum2, gpu_pixels, gpu_x_weighted, gpu_y_weighted, gpu_max);
|
|
|
|
cudaMemcpyAsync(host_sum.data(), gpu_sum, sizeof(unsigned long long) * roi_count, cudaMemcpyDeviceToHost, *stream);
|
|
cudaMemcpyAsync(host_sum2.data(), gpu_sum2, sizeof(unsigned long long) * roi_count, cudaMemcpyDeviceToHost, *stream);
|
|
cudaMemcpyAsync(host_pixels.data(), gpu_pixels, sizeof(unsigned long long) * roi_count, cudaMemcpyDeviceToHost, *stream);
|
|
cudaMemcpyAsync(host_x_weighted.data(), gpu_x_weighted, sizeof(unsigned long long) * roi_count, cudaMemcpyDeviceToHost, *stream);
|
|
cudaMemcpyAsync(host_y_weighted.data(), gpu_y_weighted, sizeof(unsigned long long) * roi_count, cudaMemcpyDeviceToHost, *stream);
|
|
cudaMemcpyAsync(host_max.data(), gpu_max, sizeof(int) * roi_count, cudaMemcpyDeviceToHost, *stream);
|
|
cuda_err(cudaStreamSynchronize(*stream));
|
|
|
|
for (uint16_t r = 0; r < roi_count; r++) {
|
|
roi_sum[r] = static_cast<int64_t>(host_sum[r]);
|
|
roi_sum2[r] = host_sum2[r];
|
|
roi_pixels[r] = host_pixels[r];
|
|
roi_x_weighted[r] = static_cast<int64_t>(host_x_weighted[r]);
|
|
roi_y_weighted[r] = static_cast<int64_t>(host_y_weighted[r]);
|
|
roi_max[r] = (host_max[r] == INT_MIN) ? INT64_MIN : static_cast<int64_t>(host_max[r]);
|
|
}
|
|
Export(out);
|
|
}
|